@@ -17,11 +17,12 @@ You should have received a copy of the GNU General Public License
 along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
 """
 from __future__ import annotations
+import os
 import subprocess  # nosec B404 # nosemgrep: gitlab.bandit.B404
 from concurrent.futures import Future
 from importlib import metadata
 from types import TracebackType
-from typing import Optional, Type, Callable
+from typing import Optional, Type, Callable, Union, Tuple
 
 from gi.repository import GLib
 from proton.vpn.session import ServerList
@@ -33,24 +34,31 @@ from proton.vpn.core.api import ProtonVPNAPI, VPNAccount
 from proton.vpn.core.session_holder import ClientTypeMetadata
 from proton.vpn.core.connection import VPNConnector
 from proton.vpn.core.cache_handler import CacheHandler
+from proton.vpn.core.settings import Settings
 from proton.vpn.session.servers import LogicalServer
 from proton.vpn.session.session import FeatureFlags
 
+from proton.vpn.connection.enum import KillSwitchSetting as\
+    KillSwitchSettingEnum
+
 from proton.vpn.app.gtk.services import VPNReconnector
 from proton.vpn.app.gtk.services.reconnector.network_monitor import NetworkMonitor
 from proton.vpn.app.gtk.services.reconnector.session_monitor import SessionMonitor
 from proton.vpn.app.gtk.services.reconnector.vpn_monitor import VPNMonitor
-from proton.vpn.core.settings import Settings
+from proton.vpn.app.gtk.settings_watchers import SettingsWatchers
 from proton.vpn.app.gtk.utils import semver, glib
 from proton.vpn.app.gtk.utils.exception_handler import ExceptionHandler
 from proton.vpn.app.gtk.utils.executor import AsyncExecutor
 from proton.vpn.app.gtk.widgets.headerbar.menu.bug_report_dialog import BugReportForm
 from proton.vpn.app.gtk.config import AppConfig, APP_CONFIG
-from proton.vpn.connection.enum import KillSwitchSetting as KillSwitchSettingEnum
+from proton.vpn.app.gtk.conflicts import Conflicts, Conflict
+
+from proton.session.api import Fido2Assertion
 
 logger = logging.getLogger(__name__)
 
-WIREGUARD_PROTOCOL = "wireguard"
+DOT = "."  # pylint: disable=invalid-name
+TWO_FACTOR_AUTH_SECURITY_KEY_ENV_VARIABLE_NAME = "PROTON_VPN_U2F"
 
 
 class Controller:  # pylint: disable=too-many-public-methods, too-many-instance-attributes
@@ -89,6 +97,7 @@ class Controller:  # pylint: disable=too-many-public-methods, too-many-instance-
 
         self._app_config = app_config
         self._cache_handler = cache_handler or CacheHandler(APP_CONFIG)
+        self._settings_watchers = SettingsWatchers()
 
     async def initialize_vpn_connector(self):
         """
@@ -122,6 +131,20 @@ class Controller:  # pylint: disable=too-many-public-methods, too-many-instance-
         """
         return self.executor.submit(self._api.submit_2fa_code, code)
 
+    def generate_2fa_fido2_assertion(self) -> Future:
+        """
+        Scans for security keys and generates a FIDO2 assertion for U2F authentication.
+        :return: A Future object wrapping the Fido2Assertion.
+        """
+        return self.executor.submit(self._api.generate_2fa_fido2_assertion)
+
+    def submit_2fa_fido2(self, fido2_assertion: Fido2Assertion) -> Future:
+        """
+        Submits a 2-factor authentication using U2F for verification.
+        :return: A Future object wrapping the result of the 2FA verification.
+        """
+        return self.executor.submit(self._api.submit_2fa_fido2, fido2_assertion)
+
     def logout(self) -> Future:
         """
         Logs the user out.
@@ -142,6 +165,11 @@ class Controller:  # pylint: disable=too-many-public-methods, too-many-instance-
         """Returns user tier."""
         return self._api.user_tier
 
+    @property
+    def settings_watchers(self):
+        """Returns a registry for settings changed callbacks."""
+        return self._settings_watchers
+
     def run_startup_actions(self, _):
         """Runs any startup actions that are necessary once the app has loaded."""
         logger.info(
@@ -262,7 +290,7 @@ class Controller:  # pylint: disable=too-many-public-methods, too-many-instance-
         return self._connector.is_connection_active  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
 
     @property
-    def is_connection_disconnected(self) -> bool:
+    def connection_disconnected(self) -> bool:
         """Returns whether the current connection is in disconnected state or not."""
         return isinstance(self._connector.current_state, states.Disconnected)
 
@@ -346,7 +374,6 @@ class Controller:  # pylint: disable=too-many-public-methods, too-many-instance-
         if necessary.
         """
         async def save(settings):
-            # Save the settings to disk
             await self._api.save_settings(settings)
 
         future = self.executor.submit(
@@ -354,11 +381,113 @@ class Controller:  # pylint: disable=too-many-public-methods, too-many-instance-
             settings
         )
 
+        future.add_done_callback(
+            lambda f: GLib.idle_add(self._settings_watchers.notify, settings)
+        )
+
         if bubble_up_errors:
             glib.bubble_up_errors(future)
 
         return future
 
+    def resolve_settings_type(self, setting_path_name: str) -> Tuple[str,
+                                                                     str,
+                                                                     Callable,
+                                                                     Callable]:
+        """
+        Resolves the type of the setting based on the setting path name.
+        :param setting_path_name: The path name of the setting.
+        :return: A tuple containing the setting type, setting attributes,
+        and the methods to get and save the settings.
+        """
+        setting_type, setting_attrs = setting_path_name.split(DOT, maxsplit=1)
+        if setting_type == "settings":
+            return (
+                setting_type,
+                setting_attrs,
+                self.get_settings,
+                self.save_settings
+            )
+        if setting_type == "app_configuration":
+            return (
+                setting_type,
+                setting_attrs,
+                self.get_app_configuration,
+                self.save_app_configuration
+            )
+
+        raise ValueError(
+            f"Unknown setting type: {setting_type}. "
+            "Expected 'settings' or 'app_config'."
+        )
+
+    def setting_attr_has_conflict(self, setting_path_name: str,
+                                  new_value: object) -> Optional[Conflict]:
+        """
+        Checks if the setting can be changed without conflicts.
+        Returns a string with the conflicts if there are any, otherwise returns
+        an empty string.
+        """
+
+        setting_type, setting_attrs, get_settings, _ =\
+            self.resolve_settings_type(setting_path_name)
+
+        return Conflicts.detect(setting_type, setting_attrs, new_value,
+                                get_settings())
+
+    def get_setting_attr(self, setting_path_name: str) -> object:
+        """Helper method to get the settings.
+
+        In this case the setting_path_name can be a multi-layered setting, for example
+        the kill switch can be access via `settings.killswtich` while netshield can be accessed
+        via `settings.features.netshield`. Since this method tries to abstract the depth,
+        we'll be searching based on the hierarchy, example:
+
+        ```
+        setting_path_name = "settings.features.netshield"
+        setting_type = "settings"
+
+        # The below will become ["features", "netshield"]
+        setting_attrs_split = "features.netshield".split(".")
+        ```
+        So the for loop will loop for each attribute and attempt to get it from the original
+        settings object, thus solving the nesting situation.
+
+        Args:
+            setting_path_name (str):
+
+        Returns:
+            object:
+        """
+        _, setting_attrs, get_settings, _ =\
+            self.resolve_settings_type(setting_path_name)
+
+        settings = get_settings()
+        for attr in setting_attrs.split(DOT):
+            settings = getattr(settings, attr)
+
+        return settings
+
+    def save_setting_attr(self, setting_path_name: str, new_value: object):
+        """Helper method to save the settings."""
+        def set_setting(root, attr, value):
+            if attr.count(DOT) == 0:
+                setattr(root, attr, value)
+            else:
+                name, path = attr.split(DOT, maxsplit=1)
+                next_root = getattr(root, name)
+                set_setting(next_root, path, value)
+
+        setting_type, setting_attrs, get_settings, save_settings =\
+            self.resolve_settings_type(setting_path_name)
+
+        settings = get_settings()
+
+        set_setting(settings, setting_attrs, new_value)
+
+        save_settings(Conflicts.resolve(setting_type, setting_attrs, new_value,
+                                        settings))
+
     def get_available_protocols(self) -> Optional[str]:
         """Returns an alphabetically sorted list of available protocol to use."""
         available_protocols = self._connector.get_available_protocols_for_backend(
@@ -370,11 +499,16 @@ class Controller:  # pylint: disable=too-many-public-methods, too-many-instance-
             key=lambda protocol: protocol.cls.ui_protocol
         )
 
-    def send_error_to_proton(self,
-                             error: BaseException |
-                             tuple[Optional[Type[BaseException]],
-                                   Optional[BaseException],
-                                   Optional[TracebackType]]):
+    def send_error_to_proton(
+        self,
+        error: Union[
+            BaseException,
+            tuple[
+                Optional[Type[BaseException]],
+                Optional[BaseException],
+                Optional[TracebackType]]
+            ]
+    ):
         """Sends the error to Sentry."""
         self._api.usage_reporting.report_error(error)
 
@@ -451,3 +585,27 @@ class Controller:  # pylint: disable=too-many-public-methods, too-many-instance-
         """Unsets the callback that is called when the server loads are updated."""
         future = self.executor.submit(self._api.refresher.set_server_loads_updated_callback, None)
         future.add_done_callback(lambda f: GLib.idle_add(f.result))
+
+    @property
+    def split_tunneling_available(self) -> bool:
+        """Returns if split tunneling is available.
+
+        Returns:
+            bool: `True` if available, `False` otherwise
+        """
+        return self._connector.is_split_tunneling_available
+
+    @property
+    def fido2_available(self) -> bool:
+        """
+        Returns if FIDO2 is available.
+        """
+        return self._api.is_fido2_lib_available
+
+    @property
+    def security_key_env_variable_set(self) -> bool:
+        """Returns if the environment variable is set for security key 2FA."""
+        env_variable_value = os.environ.get(
+            TWO_FACTOR_AUTH_SECURITY_KEY_ENV_VARIABLE_NAME, "").lower()
+
+        return env_variable_value in ["1", "true", "yes", "y"]
