@@ -19,18 +19,17 @@ GNU General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
 """
-from typing import List, Tuple, Callable, Union
+from typing import List, Tuple, Callable, Any
 from gi.repository import Gtk, Gdk
 from proton.vpn.app.gtk.controller import Controller
+from proton.vpn.app.gtk.widgets.main.confirmation_dialog \
+    import ConfirmationDialog, show_confirmation_dialog
 
 
 RECONNECT_MESSAGE = "Please establish a new VPN connection for "\
         "changes to take effect."
 
 
-DOT = "."  # pylint: disable=invalid-name
-
-
 class CategoryHeader(Gtk.Label):
     """Header that is used to seperate between setting types, such as
     Features, Connection, etc."""
@@ -41,6 +40,32 @@ class CategoryHeader(Gtk.Label):
         style_context.add_class("heading")
 
 
+class ReactiveSetting:  # pylint: disable=too-few-public-methods
+    """Base class for reactive settings that need to be updated when settings change."""
+    def on_settings_changed(self, _settings):
+        """Method that is called when settings are changed.
+        This is used to update the widget when settings change."""
+        raise NotImplementedError("This method should be implemented in subclasses.")
+
+
+class ReactiveSettingContainer:  # pylint: disable=too-few-public-methods
+    """Base class for containers that hold reactive settings."""
+    def get_children(self):
+        """Returns the children of this container, normally implemented by
+        the Gtk.Container class, it must be implemented in order
+        for this class to work."""
+
+        raise NotImplementedError(
+            "This method should be implemented.")
+
+    def on_settings_changed(self, settings):
+        """Method that is called when settings are changed.
+        This is used to update the widget when settings change."""
+        for child in self.get_children():
+            if isinstance(child, ReactiveSetting):
+                child.on_settings_changed(settings)
+
+
 class BaseCategoryContainer(Gtk.Box):
     """Base container class that is used to group common settings.
 
@@ -121,49 +146,6 @@ def is_upgrade_required(requires_subscription_to_be_active: bool, user_tier: int
     return requires_subscription_to_be_active and user_tier < 1
 
 
-def get_setting(controller: Controller, setting_path_name: str):
-    """Helper method to get the settings.
-
-    In this case the setting_path_name can be a multi-layered setting, for example
-    the kill switch can be access via `settings.killswtich` while netshield can be accessed
-    via `settings.features.netshield`. Since this method tries to abstract the depth,
-    we'll be searching based on the hierarchy, example:
-    ```
-    setting_path_name = "settings.features.netshield"
-    setting_type = "settings"
-    setting_attrs_split = "features.netshield".split(".") # <- will become ["features", "netshield"]
-    ```
-    So the for loop will loop for each attribute and attempt to get it from the original
-    settings object, thus solving the nesting situation.
-    """
-    setting_type, setting_attrs = setting_path_name.split(DOT, maxsplit=1)
-    settings = getattr(controller, f"get_{setting_type}")()
-    setting_attrs_split = setting_attrs.split(DOT)
-
-    for attr in setting_attrs_split:
-        settings = getattr(settings, attr)
-
-    return settings
-
-
-def save_setting(controller: Controller, setting_path_name: str, new_value: Union[str, int]):
-    """Helper method to save the settings."""
-    def set_setting(root, attr, value):
-        if attr.count(DOT) == 0:
-            setattr(root, attr, value)
-        else:
-            name, path = attr.split(DOT, maxsplit=1)
-            set_setting(getattr(root, name), path, value)
-
-    setting_type, setting_attrs = setting_path_name.split(DOT, maxsplit=1)
-
-    save_settings_method = getattr(controller, f"save_{setting_type}")
-    settings = getattr(controller, f"get_{setting_type}")()
-    set_setting(settings, setting_attrs, new_value)
-
-    save_settings_method(settings)
-
-
 class CustomButton(Gtk.Grid):
     """Custom button setting."""
     def __init__(  # pylint: disable=too-many-arguments
@@ -222,7 +204,9 @@ class ToggleWidget(Gtk.Grid):  # pylint: disable=too-many-instance-attributes
         setting_name: str,
         requires_subscription_to_be_active: bool = False,
         callback: Callable = None,
-        disable_on_active_connection: bool = False
+        disable_on_active_connection: bool = False,
+        enabled: bool = None,
+        display_tooltip_only_on_active_connection: bool = False
     ):
         super().__init__()
         self._apply_grid_styles()
@@ -231,9 +215,13 @@ class ToggleWidget(Gtk.Grid):  # pylint: disable=too-many-instance-attributes
         self._callback = callback
         self._requires_subscription_to_be_active = requires_subscription_to_be_active
         self._disable_on_active_connection = disable_on_active_connection
+        self._enabled = enabled
+        self._display_tooltip_only_on_active_connection = display_tooltip_only_on_active_connection
+
         self.label = SettingName(title)
         self.description = SettingDescription(description)
         self.switch = self._build_switch()
+
         self._build_ui()
 
     @property
@@ -248,11 +236,11 @@ class ToggleWidget(Gtk.Grid):  # pylint: disable=too-many-instance-attributes
 
     def get_setting(self) -> bool:
         """Shortcut property that returns the current setting"""
-        return get_setting(self._controller, self._setting_name)
+        return self._controller.get_setting_attr(self._setting_name)
 
     def save_setting(self, new_value: bool):
         """Shortcut property that sets the new setting and stores to disk."""
-        save_setting(self._controller, self._setting_name, new_value)
+        self._controller.save_setting_attr(self._setting_name, new_value)
 
     @property
     def overridden_by_upgrade_tag(self) -> bool:
@@ -262,8 +250,12 @@ class ToggleWidget(Gtk.Grid):  # pylint: disable=too-many-instance-attributes
 
     def set_tooltip(self, tooltip_text: str):
         """Set a tooltip to this row."""
-        self.set_has_tooltip(True)
-        self.set_tooltip_text(tooltip_text)
+        if (
+            self._display_tooltip_only_on_active_connection
+            and not self._controller.connection_disconnected
+        ):
+            self.set_has_tooltip(True)
+            self.set_tooltip_text(tooltip_text)
 
     def _apply_grid_styles(self):
         self.get_style_context().add_class("setting-item")
@@ -273,17 +265,17 @@ class ToggleWidget(Gtk.Grid):  # pylint: disable=too-many-instance-attributes
 
     def _build_switch(self) -> Gtk.Switch:
         switch = Gtk.Switch()
-        switch.set_state(self.get_setting())
-        if self._callback:
-            switch.connect("state-set", self._callback, self)
-        else:
-            switch.connect("state-set", self._on_switch_state)
+        if self._enabled is None:
+            self._enabled = self.get_setting()
+        switch.set_state(self._enabled)
+
+        switch.connect("notify::active", self._on_switch_state)
+
         return switch
 
     def _build_ui(self):
         """Builds the UI depending if an upgrade is required or not."""
         if self._is_upgrade_required:
-            self.save_setting(False)
             self.switch = UpgradePlusTag()
 
         self.attach(self.label, 0, 0, 1, 1)
@@ -297,12 +289,15 @@ class ToggleWidget(Gtk.Grid):  # pylint: disable=too-many-instance-attributes
         if self.description:
             self.attach(self.description, 0, 1, 2, 1)
 
-        if (not self._controller.is_connection_disconnected  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
-                and self._disable_on_active_connection):
+        if not self._controller.connection_disconnected and self._disable_on_active_connection:
             self.active = False
 
-    def _on_switch_state(self, _, new_value: bool):
-        self.save_setting(new_value)
+    def _on_switch_state(self, switch, _gparam):
+        new_value = switch.get_state()
+        if self._callback:
+            self._callback(self, new_value, self)
+        else:
+            self.save_setting(new_value)
 
     @property
     def _is_upgrade_required(self) -> bool:
@@ -317,6 +312,84 @@ class ToggleWidget(Gtk.Grid):  # pylint: disable=too-many-instance-attributes
         self.switch.set_state(False)
 
 
+class ConflictableToggleWidget(ToggleWidget):  # pylint: disable=too-many-instance-attributes
+    """
+    Toggle widget that can handle conflicts when toggling the switch.
+    It will show a confirmation dialog if the toggle is in conflict with
+    other settings.
+    """
+    def __init__(  # pylint: disable=too-many-arguments
+        self,
+        controller: Controller,
+        title: str,
+        description: str,
+        setting_name: str,
+        do_set: Callable[[ToggleWidget, bool], None],
+        do_revert: Callable[[ToggleWidget], None],
+        requires_subscription: bool = False,
+        disable_on_active_connection: bool = False,
+        enabled: bool = None,
+        display_tooltip_only_on_active_connection: bool = False,
+        conflict_resolver: Callable[[str, Any], str] = None,
+    ):
+        super().__init__(
+            controller=controller, title=title,
+            description=description, setting_name=setting_name,
+            requires_subscription_to_be_active=requires_subscription,
+            callback=self._on_switch_button_toggle,
+            disable_on_active_connection=disable_on_active_connection,
+            enabled=enabled,
+            display_tooltip_only_on_active_connection=display_tooltip_only_on_active_connection
+        )
+        self.do_set = do_set
+        self.do_revert = do_revert
+
+        # Allow to pass a custom conflict resolver
+        if conflict_resolver is None:
+            conflict_resolver = controller.setting_attr_has_conflict
+        self._conflict_resolver = conflict_resolver
+
+    def _on_switch_button_toggle(self, _, new_value: bool, __):
+
+        if conflict := self._conflict_resolver(self._setting_name, new_value):
+
+            def confirm_change(dialog: ConfirmationDialog, response: int):
+                if response == Gtk.ResponseType.YES:
+                    # do_set can be any callable that takes
+                    # (ToggleWidget, int) as arguments so although it looks
+                    # like we are calling a method on self, we are actually
+                    # calling the method that was passed to the constructor.
+                    # This allows for instantiating this class with different
+                    # do_set and do_revert methods without having to
+                    # subclass it.
+                    #
+                    # We are passing `self` as the first argument
+                    # because the do_set method expects a ToggleWidget.
+                    self.do_set(self, new_value)
+                else:
+                    # Similarly to above, we are calling the
+                    # do_revert method that was passed to the constructor,
+                    # it's not a method of this class.
+                    # This is why we are passing `self` as the first argument.
+                    self.do_revert(self)
+
+                # We cant just close the dialog, instead we destroy it
+                # directly.
+                dialog.destroy()
+
+            show_confirmation_dialog(
+                self.get_toplevel(),
+                title="",
+                question=conflict.label,
+                clarification=conflict.description,
+                yes_text="_Yes",
+                no_text="_Cancel",
+                callback_result=confirm_change
+            )
+        else:
+            self.do_set(self, new_value)
+
+
 class ComboboxWidget(Gtk.Grid):  # pylint: disable=too-many-instance-attributes
     """Default combobox text widget."""
     def __init__(  # pylint: disable=too-many-arguments
@@ -355,11 +428,11 @@ class ComboboxWidget(Gtk.Grid):  # pylint: disable=too-many-instance-attributes
 
     def get_setting(self) -> str:
         """Shortcut property that returns the current setting"""
-        return str(get_setting(self._controller, self._setting_name))
+        return str(self._controller.get_setting_attr(self._setting_name))
 
-    def save_setting(self, new_value: Union[str, int]):
+    def save_setting(self, new_value: int):
         """Shortcut property that sets the new setting and stores to disk."""
-        save_setting(self._controller, self._setting_name, new_value)
+        self._controller.save_setting_attr(self._setting_name, new_value)
 
     @property
     def overridden_by_upgrade_tag(self) -> bool:
@@ -391,8 +464,7 @@ class ComboboxWidget(Gtk.Grid):  # pylint: disable=too-many-instance-attributes
         else:
             combobox.connect("changed", self._on_combobox_change)
 
-        if (not self._controller.is_connection_disconnected  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
-                and self._disable_on_active_connection):
+        if not self._controller.connection_disconnected and self._disable_on_active_connection:
             self.active = False
 
         return combobox
@@ -432,6 +504,78 @@ class ComboboxWidget(Gtk.Grid):  # pylint: disable=too-many-instance-attributes
         self.combobox.set_active_id(str(0))
 
 
+class ConflictableComboboxWidget(ComboboxWidget):
+    """
+    Combobox widget that can handle conflicts when changing the combobox value.
+    It will show a confirmation dialog if the combobox change is in conflict with
+    other settings.
+    """
+    def __init__(  # pylint: disable=too-many-arguments
+        self,
+        controller: Controller,
+        title: str,
+        setting_name: str,
+        combobox_options: List[Tuple[int, str]],
+        do_set: Callable[[ComboboxWidget, int], None],
+        do_revert: Callable[[ComboboxWidget], None],
+        description: str = None,
+        requires_subscription: bool = False,
+        disable_on_active_connection: bool = False
+    ):
+        super().__init__(
+            controller=controller, title=title, setting_name=setting_name,
+            combobox_options=combobox_options, description=description,
+            requires_subscription_to_be_active=requires_subscription,
+            callback=self._on_conflictable_combobox_change,
+            disable_on_active_connection=disable_on_active_connection
+        )
+        self._do_set = do_set
+        self._do_revert = do_revert
+
+    def _on_conflictable_combobox_change(
+            self, combobox_text_widget, _combobox: Gtk.ComboBox):
+        new_value = combobox_text_widget.get_active_text()
+
+        if conflict := self._controller.setting_attr_has_conflict(
+                self._setting_name, new_value):
+
+            def confirm_change(dialog: ConfirmationDialog, response: int):
+                if response == Gtk.ResponseType.YES:
+                    # do_set can be any callable that takes
+                    # (ComboboxWidget, int) as arguments so although it looks
+                    # like we are calling a method on self, we are actually
+                    # calling the method that was passed to the constructor.
+                    # This allows for instantiating this class with different
+                    # do_set and do_revert methods without having to
+                    # subclass it.
+                    #
+                    # We are passing `self` as the first argument
+                    # because the do_set method expects a ComboboxWidget.
+                    self._do_set(self, new_value)
+                else:
+                    # Similarly to above, we are calling the
+                    # do_revert method that was passed to the constructor,
+                    # it's not a method of this class.
+                    # This is why we are passing `self` as the first argument.
+                    self._do_revert(self)
+
+                # We cant just close the dialog, instead we destroy it
+                # directly.
+                dialog.destroy()
+
+            show_confirmation_dialog(
+                self.get_toplevel(),
+                title="",
+                question=conflict.label,
+                clarification=conflict.description,
+                yes_text="_Yes",
+                no_text="_Cancel",
+                callback_result=confirm_change
+            )
+        else:
+            self._do_set(self, new_value)
+
+
 class EntryWidget(Gtk.Grid):
     """Default entry widget."""
     def __init__(  # pylint: disable=too-many-arguments
@@ -464,13 +608,13 @@ class EntryWidget(Gtk.Grid):
         """Set if the widget should be active or not."""
         self.set_property("sensitive", new_value)
 
-    def get_setting(self) -> Union[str, List[str]]:
+    def get_setting(self) -> bool:
         """Shortcut property that returns the current setting"""
-        return get_setting(self._controller, self._setting_name)
+        return self._controller.get_setting_attr(self._setting_name)
 
-    def save_setting(self, new_value: str):
+    def save_setting(self, new_value: bool):
         """Shortcut property that sets the new setting and stores to disk."""
-        save_setting(self._controller, self._setting_name, new_value)
+        self._controller.save_setting_attr(self._setting_name, new_value)
 
     @property
     def overridden_by_upgrade_tag(self) -> bool:
