@@ -26,137 +26,19 @@ from proton.vpn.app.gtk.widgets.main.confirmation_dialog import ConfirmationDial
 from proton.vpn.core.settings import NetShield
 from proton.vpn.app.gtk.controller import Controller
 from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import (
-    BaseCategoryContainer, ComboboxWidget, ToggleWidget, SettingName, SettingDescription
+    BaseCategoryContainer, ComboboxWidget, ToggleWidget,
+    ReactiveSettingContainer
 )
-from proton.vpn.connection.enum import KillSwitchSetting as KillSwitchSettingEnum
 from proton.vpn.app.gtk.widgets.headerbar.menu.settings.custom_dns import CustomDNSWidget
+from proton.vpn.app.gtk.widgets.headerbar.menu.settings.kill_switch import KillSwitchWidget
+from proton.vpn.app.gtk.widgets.headerbar.menu.settings.split_tunneling import SplitTunnelingToggle
 
 if TYPE_CHECKING:
     from proton.vpn.app.gtk.widgets.headerbar.menu.settings.settings_window import \
         SettingsWindow
 
 
-class KillSwitchWidget(ToggleWidget):  # noqa pylint: disable=too-many-instance-attributes,too-few-public-methods
-    """Kill switch setting widget.
-
-    Since the kill switch can have multiple modes, we need to have a proper
-    widget that handles of all of these cases and is easy to test.
-    """
-    KILLSWITCH_LABEL = "Kill switch"
-    KILLSWITCH_DESCRIPTION = "Protects your IP address by disconnecting you from the " \
-        "internet if you lose your VPN connection. "\
-        "<a href=\"https://protonvpn.com/support/what-is-kill-switch/\">Learn more</a> \n\n" \
-        "Kill switch can only be changed when VPN is disconnected."
-    KILLSWITCH_STANDARD_DESCRIPTION = "Automatically disconnect from the internet if "\
-        "VPN connection is lost."
-    KILLSWITCH_ADVANCED_DESCRIPTION = "Only allow internet access when connected to Proton VPN. " \
-        "Advanced kill switch will remain active even when you restart your device."
-    SETTING_NAME = "settings.killswitch"
-
-    def __init__(self, controller: Controller, gtk: Gtk = None):
-        super().__init__(
-            controller=controller,
-            title=self.KILLSWITCH_LABEL,
-            description=self.KILLSWITCH_DESCRIPTION,
-            setting_name=self.SETTING_NAME,
-            callback=self._on_switch_button_toggle,
-            disable_on_active_connection=True
-        )
-
-        self.gtk = gtk or Gtk
-        self._controller = controller
-        self._standard_radio_button_connect_id = None
-        self._advanced_radio_button_connect_id = None
-
-        self.standard_radio_button = None
-        self.advanced_radio_button = None
-        self.revealer = None
-
-    def build_revealer(self):
-        """Builds the revealer"""
-        self.revealer = self.gtk.Revealer()
-        self.attach(self.revealer, 0, 2, 2, 1)
-        revealer_container = self._build_revealer_container()
-        self.revealer.add(revealer_container)
-        self.revealer.set_reveal_child(self.get_setting() > KillSwitchSettingEnum.OFF)
-
-    @staticmethod
-    def build(controller: Controller) -> "KillSwitchWidget":
-        """Shortcut method to initialize widget."""
-        widget = KillSwitchWidget(controller)
-        widget.build_revealer()
-        return widget
-
-    def _build_revealer_container(self) -> Gtk.Box:
-        # Add both containers that contain all children that are to be displayed in the revealer
-        revealer_container = self.gtk.Box(orientation=Gtk.Orientation.VERTICAL)
-        revealer_container.set_spacing(10)
-        revealer_container.pack_start(self._build_standard_killswitch(), False, False, 0)
-        revealer_container.pack_start(self._build_advanced_killswitch(), False, False, 0)
-
-        return revealer_container
-
-    def _build_standard_killswitch(self) -> Gtk.Grid:
-        main_standard_container = self.gtk.Grid()
-        main_standard_container.set_column_spacing(10)
-
-        self.standard_radio_button = self.gtk.RadioButton()
-        self.standard_radio_button.set_active(self.get_setting() == KillSwitchSettingEnum.ON)
-
-        main_standard_container.attach(self.standard_radio_button, 0, 0, 1, 1)
-        main_standard_container.attach(SettingName("Standard"), 1, 0, 1, 1)
-        main_standard_container.attach(
-            SettingDescription(self.KILLSWITCH_STANDARD_DESCRIPTION),
-            1, 1, 1, 1
-        )
-
-        self.standard_radio_button.connect(
-            "toggled", self._on_radio_button_toggle, KillSwitchSettingEnum.ON
-        )
-
-        return main_standard_container
-
-    def _build_advanced_killswitch(self) -> Gtk.Grid:
-        main_advanced_container = self.gtk.Grid()
-        main_advanced_container.set_column_spacing(10)
-
-        self.advanced_radio_button = self.gtk.RadioButton(group=self.standard_radio_button)
-        self.advanced_radio_button.set_active(self.get_setting() == KillSwitchSettingEnum.PERMANENT)
-
-        main_advanced_container.attach(self.advanced_radio_button, 0, 0, 1, 1)
-        main_advanced_container.attach(SettingName("Advanced"), 1, 0, 1, 1)
-        main_advanced_container.attach(
-            SettingDescription(self.KILLSWITCH_ADVANCED_DESCRIPTION),
-            1, 1, 1, 1
-        )
-
-        self.advanced_radio_button.connect(
-            "toggled", self._on_radio_button_toggle, KillSwitchSettingEnum.PERMANENT
-        )
-
-        return main_advanced_container
-
-    def _on_radio_button_toggle(self, radio_button: Gtk.RadioButton, new_value: int):
-        # If revealer is hidden then we don't want to resolve the trigger from
-        # programmatically setting the standard radio button.
-        if not self.revealer.get_reveal_child():
-            return
-
-        if radio_button.get_active():
-            self._update(new_value, False)
-
-    def _on_switch_button_toggle(self, _, new_value: bool, __):
-        self._update(int(new_value), True)
-
-    def _update(self, value: int, new_value_comes_from_main_switch: bool):
-        self.save_setting(value)
-        self.revealer.set_reveal_child(value > KillSwitchSettingEnum.OFF)
-
-        if new_value_comes_from_main_switch:
-            self.standard_radio_button.set_active(True)
-
-
-class FeatureSettings(BaseCategoryContainer):  # pylint: disable=too-many-instance-attributes
+class FeatureSettings(BaseCategoryContainer, ReactiveSettingContainer):  # noqa: E501 # pylint: disable=line-too-long, too-many-instance-attributes
     """Settings related to connection are all grouped under this class."""
     CATEGORY_NAME = "Features"
     NETSHIELD_LABEL = "NetShield"
@@ -165,6 +47,9 @@ class FeatureSettings(BaseCategoryContainer):  # pylint: disable=too-many-instan
     PORT_FORWARDING_LABEL = "Port forwarding"
     PORT_FORWARDING_DESCRIPTION = "Bypass firewalls to connect to P2P servers "\
         "and devices on your local network."
+    PORT_FORWARDING_DESCRIPTION_LEARN_MORE = "Bypass firewalls to connect to P2P servers "\
+        "and devices on your local network. "\
+        "<a href=\"https://protonvpn.com/support/port-forwarding/#linux\">Learn more</a>"
     PORT_FORWARDING_SETUP_GUIDE = "Follow our "\
         "<a href=\"https://protonvpn.com/support/port-forwarding-manual-setup/"\
         "#how-to-use-port-forwarding\">guide</a>"\
@@ -184,6 +69,11 @@ class FeatureSettings(BaseCategoryContainer):  # pylint: disable=too-many-instan
         self.build_netshield()
         self.build_killswitch()
         self.build_port_forwarding()
+        if (
+            self._controller.split_tunneling_available
+            and self._controller.feature_flags.get("DisplaySplitTunneling")
+        ):
+            self.build_split_tunneling()
 
     def build_netshield(self):
         """Builds and adds the `netshield` setting to the widget.
@@ -196,7 +86,6 @@ class FeatureSettings(BaseCategoryContainer):  # pylint: disable=too-many-instan
             treeiter = combobox.get_active_iter()
             netshield = int(model[treeiter][1])
             combobox_widget.save_setting(netshield)
-            self._settings_window.notify_user_with_reconnect_message()
             self.emit("netshield-setting-changed", netshield)
 
         netshield_options = [
@@ -217,24 +106,27 @@ class FeatureSettings(BaseCategoryContainer):  # pylint: disable=too-many-instan
 
     def build_killswitch(self):
         """Builds and adds the `killswitch` setting to the widget."""
-        killswitch = KillSwitchWidget.build(self._controller)
-        self.pack_start(killswitch, False, False, 0)
+        self.pack_start(
+            KillSwitchWidget.build(self._controller), False, False, 0)
 
     def build_port_forwarding(self):
         """Builds and adds the `port_forwarding` setting to the widget."""
-        def on_switch_state(_, new_value: bool, toggle_widget: ToggleWidget):
-            description_value = self.PORT_FORWARDING_DESCRIPTION
+
+        def on_switch_state(_, enabled: bool, toggle_widget: ToggleWidget):
 
             # When we start displaying port forwarding, we no longer want to be showing the
             # setup guide.
-            if new_value and not self._controller.feature_flags.get("DisplayPortForwarding"):
+            display_port_forwarding = self._controller.feature_flags.get("DisplayPortForwarding")
+
+            description_value = self.PORT_FORWARDING_DESCRIPTION
+            if enabled:
                 description_value = self.PORT_FORWARDING_SETUP_GUIDE
+                if display_port_forwarding:
+                    description_value = self.PORT_FORWARDING_DESCRIPTION_LEARN_MORE
 
-            toggle_widget.save_setting(new_value)
+            toggle_widget.save_setting(enabled)
             toggle_widget.description.set_label(description_value)
 
-            self._settings_window.notify_user_with_reconnect_message()
-
         port_forwarding_widget = ToggleWidget(
             controller=self._controller,
             title=self.PORT_FORWARDING_LABEL,
@@ -243,8 +135,14 @@ class FeatureSettings(BaseCategoryContainer):  # pylint: disable=too-many-instan
             requires_subscription_to_be_active=True,
             callback=on_switch_state
         )
-        if port_forwarding_widget.get_setting():
-            port_forwarding_widget.description.set_label(self.PORT_FORWARDING_SETUP_GUIDE)
+        is_pf_enabled = port_forwarding_widget.get_setting()
+        display_port_forwarding = self._controller.feature_flags.get("DisplayPortForwarding")
+        if is_pf_enabled:
+            port_forwarding_widget.description.set_label(
+                self.PORT_FORWARDING_DESCRIPTION_LEARN_MORE
+                if display_port_forwarding
+                else self.PORT_FORWARDING_SETUP_GUIDE
+            )
 
         self.pack_start(port_forwarding_widget, False, False, 0)
 
@@ -310,3 +208,10 @@ class FeatureSettings(BaseCategoryContainer):  # pylint: disable=too-many-instan
         container.pack_start(learn_more, False, False, 0)
 
         return container
+
+    def build_split_tunneling(self):
+        """Build split tunneling UI.
+        """
+        self.pack_start(
+            SplitTunnelingToggle.build(self._controller), False, False, 0
+        )
